unit EditLangUnit;

interface

{$I CnPack.inc}

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  FileCtrl, CnCommon, ExtCtrls, StdCtrls, ComCtrls, Grids, ToolWin, ImgList,
  ActnList;

type
  TEditLangForm = class(TForm)
    statMain: TStatusBar;
    pnlTop: TPanel;
    pnlLeftTop: TPanel;
    cbbLeftDir: TComboBox;
    cbbLeftFile: TComboBox;
    spl1: TSplitter;
    pnlRightTop: TPanel;
    cbbRightDir: TComboBox;
    cbbRightFile: TComboBox;
    StringGrid: TStringGrid;
    tlbEdit: TToolBar;
    actlstMain: TActionList;
    actNextDiff: TAction;
    actPrevDiff: TAction;
    actSaveLeft: TAction;
    actSaveRight: TAction;
    ilMain: TImageList;
    btnNextDiff: TToolButton;
    btnPrevDiff: TToolButton;
    btn1: TToolButton;
    btnSaveLeft: TToolButton;
    btnSaveRight: TToolButton;
    btn2: TToolButton;
    actTSortLeft: TAction;
    actTSortRight: TAction;
    btnTSortLeft: TToolButton;
    btnTSortRight: TToolButton;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure cbbLeftDirChange(Sender: TObject);
    procedure cbbRightDirChange(Sender: TObject);
    procedure cbbLeftFileChange(Sender: TObject);
    procedure cbbRightFileChange(Sender: TObject);
    procedure spl1Moved(Sender: TObject);
    procedure actNextDiffExecute(Sender: TObject);
    procedure actPrevDiffExecute(Sender: TObject);
    procedure actSaveLeftExecute(Sender: TObject);
    procedure actSaveRightExecute(Sender: TObject);
    procedure StringGridSelectCell(Sender: TObject; ACol, ARow: Integer;
      var CanSelect: Boolean);
    procedure actTSortLeftExecute(Sender: TObject);
    procedure actTSortRightExecute(Sender: TObject);
  private
    FLangRoot: string;
    FLangDirs: TStrings;
    FLeftFileName: string;
    FRightFileName: string;
    FLeftLangFiles: TStrings;
    FRightLangFiles: TStrings;
    FLeftContent: TStrings;
    FRightContent: TStrings;
    FLeftDisplay: TStrings;
    FRightDisplay: TStrings;
    function SaveLinesToFile(Lines: TStrings; const FileName: string): Boolean;
    procedure SyncLeftGridToDisplay;
    procedure SyncRightGridToDisplay;
    procedure ChangeGridColumnSize;
    function LineEqual(const S1: string; const S2: string): Boolean;
    procedure SearchLangFiles(const Dir: string; List: TStrings; IsDir: Boolean);
    procedure RearrangeDisplays;
    //  LeftContent/RightContentʵʩԱȣѶŵ LeftDisplay/RightDisplay
    procedure UpdateToGrid;
    procedure StringsTSort(Lines: TStrings);
  public
    { Public declarations }
  end;

var
  EditLangForm: TEditLangForm;

implementation

{$R *.DFM}

const
  LANG_DIR = 'Lang\';

  LEFT_EDITING_COL = 1;
  RIGHT_EDITING_COL = 4;

procedure QuickSortStrings(Lines: TStrings; L, R: Integer);
var
  I, J, P: Integer;

  function CompareStrBeforeEqual(const S1, S2: string): Integer;
  var
    P: Integer;
    P1, P2: string;
  begin
    P := Pos('=', S1);
    if P > 0 then
      P1 := Copy(S1, 1, P - 1)
    else
      P1 := S1;

    P := Pos('=', S2);
    if P > 0 then
      P2 := Copy(S2, 1, P - 1)
    else
      P2 := S1;

    Result := CompareStr(P1, P2);
  end;

begin
  repeat
    I := L;
    J := R;
    P := (L + R) shr 1;
    repeat
      while CompareStrBeforeEqual(Lines[I], Lines[P]) < 0 do Inc(I);
      while CompareStrBeforeEqual(Lines[J], Lines[P]) > 0 do Dec(J);
      if I <= J then
      begin
        if I <> J then
          Lines.Exchange(I, J);
        if P = I then
          P := J
        else if P = J then
          P := I;
        Inc(I);
        Dec(J);
      end;
    until I > J;
    if L < J then QuickSortStrings(Lines, L, J);
    L := I;
  until I >= R;
end;

procedure TEditLangForm.FormCreate(Sender: TObject);
begin
{$IFNDEF UNICODE}
  ShowMessage('Must Run under Unicode Environment.');
  Application.Terminate;
{$ENDIF}

{$IFNDEF TSTRINGS_HAS_WRITEBOM}
  ShowMessage('NO BOM Support. Be Careful when Saving.');
{$ENDIF}

  FLangRoot := IncludeTrailingPathDelimiter(ExtractFileDir(Application.ExeName)) + LANG_DIR;
  FLangDirs := TStringList.Create;

  SearchLangFiles(FLangRoot, FLangDirs, True);

  cbbLeftDir.Items.Assign(FLangDirs);
  cbbRightDir.Items.Assign(FLangDirs);

  FLeftContent := TStringList.Create;
  FRightContent := TStringList.Create;
  FLeftDisplay := TStringList.Create;
  FRightDisplay := TStringList.Create;

//  mmoLeft := TCnMemo.Create(Self);
//  with mmoLeft do
//  begin
//    Align := alClient;
//    ShowLineNumber := True;
//    Parent := pnlLeft;
//    ScrollBars := ssBoth;
//  end;
//
//  mmoRight := TCnMemo.Create(Self);
//  with mmoRight do
//  begin
//    Align := alClient;
//    ShowLineNumber := True;
//    Parent := pnlRight;
//    ScrollBars := ssBoth;
//  end;
end;

procedure TEditLangForm.FormDestroy(Sender: TObject);
begin
  FLeftDisplay.Free;
  FRightDisplay.Free;
  FLeftContent.Free;
  FRightContent.Free;
  FLangDirs.Free;
  FLeftLangFiles.Free;
  FRightLangFiles.Free;
end;

function TEditLangForm.SaveLinesToFile(Lines: TStrings;
  const FileName: string): Boolean;
var
  I: Integer;
  Sl: TStringList;
begin
  Result := False;
  if (FileName <> '') and (Lines.Count > 0) then
  begin
    Sl := TStringList.Create;
    Sl.DefaultEncoding := TEncoding.UTF8;
    Sl.WriteBOM := True;
    try
      for I := 0 to Lines.Count - 1 do
        Sl.Add(Lines[I]);

      for I := Sl.Count - 1 downto 0 do
        if Sl[I] = '' then
          Sl.Delete(I);

      if Sl.Count > 0 then
      begin
        Sl.SaveToFile(FileName);
        Result := True;
      end;
    finally
      Sl.Free;
    end;
  end;
end;

procedure TEditLangForm.SearchLangFiles(const Dir: string; List: TStrings; IsDir: Boolean);
var
  SearchRec: TSearchRec;
  F: Integer;
begin
  F := FindFirst(IncludeTrailingPathDelimiter(Dir) + '*.*', faAnyFile, SearchRec);
  List.Clear;
  while F = 0 do
  begin
    if (SearchRec.Name <> '.') and (SearchRec.name <> '..') then
    begin
      if not IsDir and (SearchRec.Attr and faDirectory <> faDirectory) then
        List.Add(ExtractFileName(SearchRec.Name))
      else if IsDir and (SearchRec.Attr and faDirectory = faDirectory) then
        List.Add(ExtractFileName(SearchRec.Name));
    end;
    F := FindNext(SearchRec);
  end;
  FindClose(SearchRec);
end;

procedure TEditLangForm.spl1Moved(Sender: TObject);
begin
  ChangeGridColumnSize;
end;

procedure TEditLangForm.StringGridSelectCell(Sender: TObject; ACol,
  ARow: Integer; var CanSelect: Boolean);
begin
  if ACol in [LEFT_EDITING_COL, RIGHT_EDITING_COL] then
    StringGrid.Options := StringGrid.Options + [goEditing]
  else
    StringGrid.Options := StringGrid.Options - [goEditing];
end;

procedure TEditLangForm.StringsTSort(Lines: TStrings);
var
  I, L, R, P: Integer;
  S, OldPre, Pre: string;
begin
  if Lines.Count < 2 then
    Exit;

  I := 0;
  L := 0;
  R := 0;
  OldPre := '';
  while I < Lines.Count do
  begin
    S := Lines[I];
    if S = '' then
    begin
      Inc(I);
      Continue;
    end;

    if S[1] <> 'T' then
    begin
      Inc(I);
      Continue;
    end;

    P := Pos('=', S);
    if P <= 0 then
    begin
      Inc(I);
      Continue;
    end;

    P := Pos('.', S);
    if P <= 0 then
    begin
      Inc(I);
      Continue;
    end;

    Pre := Copy(S, 1, P - 1);
    if OldPre = '' then
    begin
      OldPre := Pre;
      L := I;
      R := I;
      Continue;
    end
    else if OldPre = Pre then
    begin
      Inc(R);
      Inc(I);
      Continue;
    end
    else if OldPre <> Pre then
    begin
      QuickSortStrings(Lines, L, R);
      OldPre := Pre;
      L := I;
      R := I;
      Inc(I);
      Continue;
    end;
  end;
end;

procedure TEditLangForm.SyncLeftGridToDisplay;
begin
  FLeftDisplay.Assign(StringGrid.Cols[LEFT_EDITING_COL]);
end;

procedure TEditLangForm.SyncRightGridToDisplay;
begin
  FRightDisplay.Assign(StringGrid.Cols[RIGHT_EDITING_COL]);
end;

procedure TEditLangForm.UpdateToGrid;
var
  I, C: Integer;
begin
  C := FLeftDisplay.Count;
  if C < FRightDisplay.Count then
    C := FRightDisplay.Count;

  StringGrid.RowCount := C;
  C := 1;
  for I := 0 to FLeftDisplay.Count - 1 do
  begin
    if FLeftDisplay[I] <> '' then
    begin
      StringGrid.Cells[0, I] := IntToStr(C);
      StringGrid.Cells[1, I] := FLeftDisplay[I];
      Inc(C);
    end
    else
    begin
      StringGrid.Cells[0, I] := '';
      StringGrid.Cells[1, I] := FLeftDisplay[I];
    end;
  end;

  C := 1;
  for I := 0 to FRightDisplay.Count - 1 do
  begin
    if FRightDisplay[I] <> '' then
    begin
      StringGrid.Cells[3, I] := IntToStr(C);
      StringGrid.Cells[4, I] := FRightDisplay[I];
      Inc(C);
    end
    else
    begin
      StringGrid.Cells[3, I] := '';
      StringGrid.Cells[4, I] := FRightDisplay[I];
    end;
  end;
end;

procedure TEditLangForm.FormResize(Sender: TObject);
const
  MARGIN = 10;
begin
  ChangeGridColumnSize;
  statMain.Panels[0].Width := Width div 2 - MARGIN;
end;

function TEditLangForm.LineEqual(const S1, S2: string): Boolean;
var
  P1, P2: Integer;
  H1, H2: string;
begin
  P1 := Pos('=', S1);
  P2 := Pos('=', S2);

  if (P1 <= 0) and (P2 <= 0) then
    Result := S1 = S2
  else if (P1 <= 0) or (P2 <= 0) then
    Result := False // һеȺһûȺ
  else // еȺ
  begin
    H1 := Copy(S1, 1, P1 - 1);
    H2 := Copy(S2, 1, P2 - 1);
    Result := H1 = H2;
  end;
end;

procedure TEditLangForm.RearrangeDisplays;
var
  I, J, L, R, LS, RS, LAC, RAC: Integer;
  Matched: Boolean;
begin
  FLeftDisplay.Clear;
  FRightDisplay.Clear;

  if FLeftContent.Count = 0 then
  begin
    FRightDisplay.Text := FRightContent.Text;
    Exit;
  end
  else if FRightContent.Count = 0 then
  begin
    FLeftDisplay.Text := FLeftContent.Text;
    Exit;
  end;

  L := 0; R := 0;
  while (L < FLeftContent.Count) and (R < FRightContent.Count) do
  begin
    if LineEqual(FLeftContent[L], FRightContent[R]) then
    begin
      FLeftDisplay.Add(FLeftContent[L]);
      FRightDisplay.Add(FRightContent[R]);
      Inc(L);
      Inc(R);
    end
    else
    begin
      LAC := 0; RAC := 0;
      // 󲻶
      RS := R + 1;
      Matched := False;
      while RS < FRightContent.Count do
      begin
        if LineEqual(FLeftContent[L], FRightContent[RS]) then
        begin
          Matched := True;
          Inc(RAC);
          Break;
        end
        else
        begin
          Inc(RS);
          Inc(RAC);
        end;
      end;

      if Matched then
      begin
        // 󲻶ұҵˣ߼RACкLұ߼RһֱӵRS
        for I := 1 to RAC do
          FLeftDisplay.Add('');
        FLeftDisplay.Add(FLeftContent[L]);
        for I := R to RS do
          FRightDisplay.Add(FRightContent[I]);

        Inc(L);
        R := RS + 1;
        Continue;
      end;

      // ûҵҲ
      LS := L + 1;
      Matched := False;
      while LS < FLeftContent.Count do
      begin
        if LineEqual(FLeftContent[LS], FRightContent[R]) then
        begin
          Matched := True;
          Inc(LAC);
          Break;
        end
        else
        begin
          Inc(LS);
          Inc(LAC);
        end;
      end;

      if Matched then
      begin
        // Ҳҵˣұ߼LACкR߼LһֱӵLS
        for I := 1 to LAC do
          FRightDisplay.Add('');
        FRightDisplay.Add(FRightContent[R]);
        for I := L to LS do
          FLeftDisplay.Add(FLeftContent[I]);

        L := LS + 1;
        Inc(R);
        Continue;
      end;

      // ûҵʵģ˵иͷˣֱ
      for I := L to FLeftContent.Count - 1 do
      begin
        FLeftDisplay.Add(FLeftContent[I]);
        FRightDisplay.Add('');
      end;
      for I := R to FRightContent.Count - 1 do
      begin
        FLeftDisplay.Add('');
        FRightDisplay.Add(FRightContent[I]);
      end;
      Break;
    end;
  end;
end;

procedure TEditLangForm.actNextDiffExecute(Sender: TObject);
var
  I: Integer;
begin
  I := StringGrid.Row + 1;
  while I < StringGrid.RowCount do
  begin
    if (FLeftDisplay[I] = '') or (FRightDisplay[I] = '') then
      Break;
    Inc(I);
  end;

  if I = StringGrid.RowCount then
    InfoDlg('No Next Different Line.')
  else
    StringGrid.Row := I;
end;

procedure TEditLangForm.actPrevDiffExecute(Sender: TObject);
var
  I: Integer;
begin
  I := StringGrid.Row - 1;
  while I >= 0 do
  begin
    if (FLeftDisplay[I] = '') or (FRightDisplay[I] = '') then
      Break;
    Dec(I);
  end;

  if I < 0 then
    InfoDlg('No Prev Different Line.')
  else
    StringGrid.Row := I;
end;

procedure TEditLangForm.actSaveLeftExecute(Sender: TObject);
begin
  if SaveLinesToFile(StringGrid.Cols[LEFT_EDITING_COL], FLeftFileName) then
    InfoDlg('Save Success. ' + FLeftFileName);
end;

procedure TEditLangForm.actSaveRightExecute(Sender: TObject);
begin
  if SaveLinesToFile(StringGrid.Cols[RIGHT_EDITING_COL], FRightFileName) then
    InfoDlg('Save Success. ' + FRightFileName);
end;

procedure TEditLangForm.actTSortLeftExecute(Sender: TObject);
begin
  StringsTSort(FLeftContent); // StringGrid.Cols[LEFT_EDITING_COL]);
  RearrangeDisplays;
  UpdateToGrid;
end;

procedure TEditLangForm.actTSortRightExecute(Sender: TObject);
begin
  StringsTSort(FRightContent); //StringGrid.Cols[RIGHT_EDITING_COL]);
  RearrangeDisplays;
  UpdateToGrid;
end;

procedure TEditLangForm.cbbLeftDirChange(Sender: TObject);
begin
  if cbbLeftDir.ItemIndex >= 0 then
    SearchLangFiles(FLangRoot + IncludeTrailingPathDelimiter(FLangDirs[cbbLeftDir.ItemIndex]),
      cbbLeftFile.Items, False);
end;

procedure TEditLangForm.cbbRightDirChange(Sender: TObject);
begin
  if cbbRightDir.ItemIndex >= 0 then
    SearchLangFiles(FLangRoot + IncludeTrailingPathDelimiter(FLangDirs[cbbRightDir.ItemIndex]),
      cbbRightFile.Items, False);
end;

procedure TEditLangForm.cbbRightFileChange(Sender: TObject);
var
  S: string;
begin
  if (cbbRightFile.Items.Count > 0) and (cbbRightFile.ItemIndex >= 0) then
  begin
    S := FLangRoot + IncludeTrailingPathDelimiter(FLangDirs[cbbRightDir.ItemIndex])
      + cbbRightFile.Items[cbbRightFile.ItemIndex];
    if FileExists(S) then
    begin
      FRightContent.LoadFromFile(S);
      FRightFileName := S;
      RearrangeDisplays;
      UpdateToGrid;
      StringGrid.SetFocus;
    end;
  end;
end;

procedure TEditLangForm.ChangeGridColumnSize;
const
  GUTTER_WIDTH = 60;
begin
  StringGrid.ColWidths[0] := GUTTER_WIDTH;
  StringGrid.ColWidths[1] := pnlLeftTop.Width - 6 - GUTTER_WIDTH;
  StringGrid.ColWidths[2] := 8;
  StringGrid.ColWidths[3] := GUTTER_WIDTH;
  StringGrid.ColWidths[4] := pnlRightTop.Width - 26 - GUTTER_WIDTH;
end;

procedure TEditLangForm.cbbLeftFileChange(Sender: TObject);
var
  S: string;
begin
  if (cbbLeftFile.Items.Count > 0) and (cbbLeftFile.ItemIndex >= 0) then
  begin
    S := FLangRoot + IncludeTrailingPathDelimiter(FLangDirs[cbbLeftDir.ItemIndex])
      + cbbLeftFile.Items[cbbLeftFile.ItemIndex];
    if FileExists(S) then
    begin
      FLeftContent.LoadFromFile(S);
      FLeftFileName := S;
      RearrangeDisplays;
      UpdateToGrid;
      StringGrid.SetFocus;
    end;
  end;
end;

end.

